from __future__ import annotations

__all__ = ["ContinuousDistributionAnalyzer", "TemporalContinuousDistributionAnalyzer"]

import logging

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import (
    ContinuousDistributionSection,
    EmptySection,
    TemporalContinuousDistributionSection,
)

logger = logging.getLogger(__name__)


class ContinuousDistributionAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    continuous values.

    Args:
    ----
        column (str): Specifies the column to analyze.
        nbins (int or None, optional): Specifies the number of bins in
            the histogram. Default: ``None``
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``
        xmin (float or str or None, optional): Specifies the minimum
            value of the range or its associated quantile.
            ``q0.1`` means the 10% quantile. ``0`` is the minimum
            value and ``1`` is the maximum value. Default: ``q0``
        xmax (float or str or None, optional): Specifies the maximum
            value of the range or its associated quantile.
            ``q0.9`` means the 90% quantile. ``0`` is the minimum
            value and ``1`` is the maximum value. Default: ``q1``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ContinuousDistributionAnalyzer
        >>> analyzer = ContinuousDistributionAnalyzer(column="float")
        >>> analyzer
        ContinuousDistributionAnalyzer(column=float, nbins=None, log_y=False, xmin=q0, xmax=q1)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        column: str,
        nbins: int | None = None,
        log_y: bool = False,
        xmin: float | str | None = "q0",
        xmax: float | str | None = "q1",
    ) -> None:
        self._column = column
        self._nbins = nbins
        self._log_y = log_y
        self._xmin = xmin
        self._xmax = xmax

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, nbins={self._nbins}, "
            f"log_y={self._log_y}, xmin={self._xmin}, xmax={self._xmax})"
        )

    def analyze(self, df: DataFrame) -> ContinuousDistributionSection | EmptySection:
        if self._column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ContinuousDistributionSection(
            column=self._column,
            series=df[self._column],
            nbins=self._nbins,
            log_y=self._log_y,
            xmin=self._xmin,
            xmax=self._xmax,
        )


class TemporalContinuousDistributionAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    continuous values.

    Args:
    ----
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import TemporalNullValueAnalyzer
        >>> analyzer = TemporalContinuousDistributionAnalyzer(
        ...     column="float", dt_column="datetime", period="M"
        ... )
        >>> analyzer
        TemporalContinuousDistributionAnalyzer(column=float, dt_column=datetime, period=M, log_y=False)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...         "datetime": pd.to_datetime(
        ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, column: str, dt_column: str, period: str, log_y: bool = False) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._log_y = log_y

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dt_column={self._dt_column}, period={self._period}, log_y={self._log_y})"
        )

    def analyze(self, df: DataFrame) -> TemporalContinuousDistributionSection | EmptySection:
        if self._column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        if self._dt_column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return TemporalContinuousDistributionSection(
            column=self._column,
            df=df,
            dt_column=self._dt_column,
            period=self._period,
            log_y=self._log_y,
        )



from __future__ import annotations

__all__ = ["SequentialPreprocessor"]

from collections.abc import Sequence

from coola.utils import str_indent, str_sequence
from pandas import DataFrame

from flamme.preprocessor.base import BasePreprocessor, setup_preprocessor


class SequentialPreprocessor(BasePreprocessor):
    r"""Implements a preprocessor to apply sequentially several
    preprocessors.

    Args:
    ----
        preprocessors (``Sequence``): Specifies the preprocessors or
            their configurations.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.preprocessor import (
        ...     SequentialPreprocessor,
        ...     StripStrPreprocessor,
        ...     ToNumericPreprocessor,
        ... )
        >>> preprocessor = SequentialPreprocessor(
        ...     [
        ...         StripStrPreprocessor(columns=["col1", "col3"]),
        ...         ToNumericPreprocessor(columns=["col1", "col2"]),
        ...     ]
        ... )
        >>> preprocessor
        SequentialPreprocessor(
          (0): StripStrPreprocessor(columns=('col1', 'col3'))
          (1): ToNumericPreprocessor(columns=('col1', 'col2'))
        )
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["a ", " b", "  c  ", "d", "e"],
        ...         "col4": ["a ", " b", "  c  ", "d", "e"],
        ...     }
        ... )
        >>> df
           col1 col2   col3   col4
        0     1    1     a      a
        1     2    2      b      b
        2     3    3    c      c
        3     4    4      d      d
        4     5    5      e      e
        >>> df = preprocessor.preprocess(df)
        >>> df
           col1 col2 col3   col4
        0     1    1    a     a
        1     2    2    b      b
        2     3    3    c    c
        3     4    4    d      d
        4     5    5    e      e
    """

    def __init__(self, preprocessors: Sequence[BasePreprocessor | dict]) -> None:
        self._preprocessors = tuple(
            setup_preprocessor(preprocessor) for preprocessor in preprocessors
        )

    def __repr__(self) -> str:
        args = ""
        if self._preprocessors:
            args = f"\n  {str_indent(str_sequence(self._preprocessors))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def preprocess(self, df: DataFrame) -> DataFrame:
        for preprocessor in self._preprocessors:
            df = preprocessor.preprocess(df)
        return df



from __future__ import annotations

__all__ = [
    "BaseSection",
    "ColumnTypeSection",
    "ContinuousDistributionSection",
    "DiscreteDistributionSection",
    "EmptySection",
    "MarkdownSection",
    "NullValueSection",
    "SectionDict",
    "TemporalContinuousDistributionSection",
    "TemporalDiscreteDistributionSection",
    "TemporalNullValueSection",
]

from flamme.section.base import BaseSection
from flamme.section.continuous import ContinuousDistributionSection
from flamme.section.continuous_temporal import TemporalContinuousDistributionSection
from flamme.section.discrete import DiscreteDistributionSection
from flamme.section.discrete_temporal import TemporalDiscreteDistributionSection
from flamme.section.dtype import ColumnTypeSection
from flamme.section.empty import EmptySection
from flamme.section.mapping import SectionDict
from flamme.section.markdown import MarkdownSection
from flamme.section.null import NullValueSection, TemporalNullValueSection



from __future__ import annotations

__all__ = ["DiscreteDistributionSection"]

from collections import Counter
from collections.abc import Sequence

import plotly
import plotly.express as px
from jinja2 import Template
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)


class DiscreteDistributionSection(BaseSection):
    r"""Implements a section that analyzes a discrete distribution of
    values.

    Args:
    ----
        counter (``Counter``): Specifies the counter that represents
            the discrete distribution.
        null_values (int): Specifies the number of null values.
        column (str, optional): Specifies the column name.
            Default: ``'N/A'``
        max_rows (int, optional): Specifies the maximum number of rows
            to show in the table. Default: ``20``
    """

    def __init__(
        self, counter: Counter, null_values: int = 0, column: str = "N/A", max_rows: int = 20
    ) -> None:
        self._counter = counter
        self._null_values = null_values
        self._column = column
        self._max_rows = int(max_rows)

        self._total = sum(self._counter.values())

    def get_statistics(self) -> dict:
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        return {
            "most_common": most_common,
            "nunique": len(most_common),
            "total": self._total,
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * self._null_values / stats['total']:.2f}" if stats["total"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "total_values": f"{stats['total']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{self._null_values:,}",
                "null_values_pct": null_values_pct,
                "figure": self._create_figure(),
                "table": self._create_table(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column {{column}}.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

{{figure}}
{{table}}
<p style="margin-top: 1rem;">
"""

    def _create_figure(self) -> str:
        if self._total == 0:
            return ""
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        df = DataFrame(
            {
                "value": [str(value) for value, _ in most_common],
                "count": [count for _, count in most_common],
            }
        )
        fig = px.bar(
            df,
            x="value",
            y="count",
            title="Number of occurrences per value",
            labels={"value": "value", "count": "number of occurrences"},
            text_auto=True,
            template="seaborn",
            log_y=(most_common[0][1] / most_common[-1][1]) >= 20,
        )
        return Template(
            r"""
<p style="margin-top: 1rem;">
<b>Distribution of values in column {{column}}</b>

<p>The values in the figure below are sorted by decreasing order of number of occurrences.

{{figure}}
"""
        ).render({"figure": plotly.io.to_html(fig, full_html=False), "column": self._column})

    def _create_table(self) -> str:
        if self._total == 0:
            return ""

        most_common = self._counter.most_common(self._max_rows)
        rows_head = "\n".join(
            [create_table_row(column=col, count=count) for col, count in most_common]
        )
        lest_common = self._counter.most_common()[-self._max_rows :][::-1]
        rows_tail = "\n".join(
            [create_table_row(column=col, count=count) for col, count in lest_common]
        )
        return Template(
            """
<details>
    <summary>Show head and tail values</summary>

    <div class="row">
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Head: {{max_values}} most common values in column {{column}}</b>
        <table class="table table-hover table-responsive w-auto" >
            <thead class="thead table-group-divider">
                <tr>
                    <th>column</th>
                    <th>count</th>
                </tr>
            </thead>
            <tbody class="tbody table-group-divider">
                {{rows_head}}
                <tr class="table-group-divider"></tr>
            </tbody>
        </table>
      </div>
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Tail: {{max_values}} least common values in column {{column}}</b>
        <table class="table table-hover table-responsive w-auto" >
            <thead class="thead table-group-divider">
                <tr>
                    <th>column</th>
                    <th>count</th>
                </tr>
            </thead>
            <tbody class="tbody table-group-divider">
                {{rows_tail}}
                <tr class="table-group-divider"></tr>
            </tbody>
        </table>
      </div>
    </div>
</details>
"""
        ).render(
            {
                "max_values": len(most_common),
                "rows_head": rows_head,
                "rows_tail": rows_tail,
                "column": self._column,
            }
        )


def create_table_row(column: str, count: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
    ----
        column (str): Specifies the column name.
        count (int): Specifies the count for the column.

    Returns:
    -------
        str: The HTML code of a row.
    """
    return Template("""<tr><th>{{column}}</th><td {{num_style}}>{{count}}</td></tr>""").render(
        {"num_style": 'style="text-align: right;"', "column": column, "count": f"{count:,}"}
    )



from __future__ import annotations

__all__ = ["PreprocessorIngestor"]

import logging

from coola.utils import str_indent, str_mapping
from pandas import DataFrame

from flamme.ingestor.base import BaseIngestor, setup_ingestor
from flamme.preprocessor.base import BasePreprocessor, setup_preprocessor

logger = logging.getLogger(__name__)


class PreprocessorIngestor(BaseIngestor):
    r"""Implements an ingestor that also preprocess the DataFrame.

    Args:
    ----
        path (``pathlib.Path`` or str): Specifies the path to the
            CSV file to ingest.
        **kwargs: Additional keyword arguments for
            ``pandas.read_csv``.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.ingestor import PreprocessorIngestor, ParquetIngestor
        >>> from flamme.preprocessor import ToNumericPreprocessor
        >>> ingestor = PreprocessorIngestor(
        ...     ingestor=ParquetIngestor(path="/path/to/df.csv"),
        ...     preprocessor=ToNumericPreprocessor(columns=["col1", "col3"]),
        ... )
        >>> ingestor
        PreprocessorIngestor(
          (ingestor): ParquetIngestor(path=/path/to/df.csv)
          (preprocessor): ToNumericPreprocessor(columns=('col1', 'col3'))
        )
        >>> df = ingestor.ingest()  # doctest: +SKIP
    """

    def __init__(
        self, ingestor: BaseIngestor | dict, preprocessor: BasePreprocessor | dict
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        self._preprocessor = setup_preprocessor(preprocessor)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"ingestor": self._ingestor, "preprocessor": self._preprocessor})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def ingest(self) -> DataFrame:
        df = self._ingestor.ingest()
        return self._preprocessor.preprocess(df)



from __future__ import annotations

__all__ = ["TemporalDiscreteDistributionSection"]

from collections.abc import Sequence

import plotly
import plotly.express as px
from jinja2 import Template
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)


class TemporalDiscreteDistributionSection(BaseSection):
    r"""Implements a section that analyzes the temporal distribution of a
    column with discrete values.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column of the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
    """

    def __init__(
        self,
        df: DataFrame,
        column: str,
        dt_column: str,
        period: str,
    ) -> None:
        self._df = df
        self._column = column
        self._dt_column = dt_column
        self._period = period

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def period(self) -> str:
        return self._period

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column {{column}} by using the column {{dt_column}}.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(df: DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a figure with the temporal value
    distribution.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``

    Returns:
    -------
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""
    df = df[[column, dt_column]].copy()
    col_dt, col_count = "__datetime__", "__count__"
    df[col_dt] = df[dt_column].dt.to_period(period).astype(str)
    # df = df[[column, col_dt]].groupby(by=col_dt, dropna=False).value_counts(dropna=False)
    df = df[[column, col_dt]].groupby(by=[col_dt, column], dropna=False)[column].size()
    df = DataFrame({col_count: df}).reset_index().sort_values(by=[col_dt, column])

    fig = px.bar(
        df,
        x=col_dt,
        y=col_count,
        color=column,
        barmode="group",
        labels={col_count: "count", col_dt: "time", column: "value"},
    )
    return plotly.io.to_html(fig, full_html=False)



from __future__ import annotations

__all__ = ["setup_object"]

import logging
from typing import TypeVar

from objectory import factory

logger = logging.getLogger(__name__)

T = TypeVar("T")


def setup_object(obj_or_config: T | dict) -> T:
    r"""Sets up an object from its configuration.

    Args:
    ----
        obj_or_config: Specifies the object or its configuration.

    Returns:
    -------
        The instantiated object.

    Example usage:

    .. code-block:: pycon

       >>> from flamme.utils import setup_object
       >>> obj = setup_object({"_target_": "collections.deque", "iterable": [1, 2, 1, 3]})
       >>> obj
       deque([1, 2, 1, 3])
       >>> setup_object(obj)  # Do nothing because the object is already instantiated
       deque([1, 2, 1, 3])
    """
    if isinstance(obj_or_config, dict):
        logger.info("Initializing an object from its configuration... ")
        obj_or_config = factory(**obj_or_config)
    return obj_or_config



from __future__ import annotations

__all__ = ["NullValueSection", "TemporalNullValueSection"]

import math
from collections.abc import Sequence

import numpy as np
import plotly
import plotly.express as px
import plotly.graph_objects as go
from jinja2 import Template
from pandas import DataFrame
from plotly.subplots import make_subplots

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)


class NullValueSection(BaseSection):
    r"""Implements a section that analyzes the number of null values.

    Args:
    ----
        columns (``Sequence``): Specifies the column names.
        null_count (``numpy.ndarray``): Specifies the number of null
            values for each column.
        total_count (``numpy.ndarray``): Specifies the total number
            of values for each column.
    """

    def __init__(
        self, columns: Sequence[str], null_count: np.ndarray, total_count: np.ndarray
    ) -> None:
        self._columns = tuple(columns)
        self._null_count = null_count.flatten().astype(int)
        self._total_count = total_count.flatten().astype(int)

        if len(self._columns) != self._null_count.shape[0]:
            raise RuntimeError(
                f"columns ({len(self._columns):,}) and null_count ({self._null_count.shape[0]:,}) "
                "do not match"
            )
        if len(self._columns) != self._total_count.shape[0]:
            raise RuntimeError(
                f"columns ({len(self._columns):,}) and total_count "
                f"({self._total_count.shape[0]:,}) do not match"
            )

    def get_statistics(self) -> dict:
        return {
            "columns": self._columns,
            "null_count": tuple(self._null_count.tolist()),
            "total_count": tuple(self._total_count.tolist()),
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table_alpha": self._create_table(sort_by="column"),
                "table_sort": self._create_table(sort_by="null"),
                "bar_figure": self._create_bar_figure(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the number/proportion of null values for each column.
The color indicates the proportion of missing values.
Dark blues indicates more missing values than light blues.

{{bar_figure}}

<div class="container-fluid">
    <div class="row align-items-start">
        <div class="col align-self-center">
            <p><b>Columns sorted by alphabetical order</b></p>

            {{table_alpha}}

        </div>
        <div class="col">
            <p><b>Columns sorted by ascending order of missing values</b></p>

            {{table_sort}}

        </div>
    </div>
</div>
"""

    def _create_bar_figure(self) -> str:
        df = self._get_dataframe().sort_values(by="null")
        fig = px.bar(
            df,
            x="column",
            y="null",
            title="number of null values per column",
            labels={"column": "column", "null": "number of null values"},
            text_auto=True,
            template="seaborn",
        )
        return plotly.io.to_html(fig, full_html=False)

    def _create_table(self, sort_by: str) -> str:
        df = self._get_dataframe().sort_values(by=sort_by)
        rows = "\n".join(
            [
                create_table_row(column=column, null_count=null_count, total_count=total_count)
                for column, null_count, total_count in zip(
                    df["column"].to_numpy(), df["null"].to_numpy(), df["total"].to_numpy()
                )
            ]
        )
        return Template(
            """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>column</th>
            <th>null pct</th>
            <th>null count</th>
            <th>total count</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
        ).render({"rows": rows})

    def _get_dataframe(self) -> DataFrame:
        return DataFrame(
            {"column": self._columns, "null": self._null_count, "total": self._total_count}
        )


def create_table_row(column: str, null_count: int, total_count: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
    ----
        column (str): Specifies the column name.
        null_count (int): Specifies the number of null values.
        total_count (int): Specifies the total number of rows.

    Returns:
    -------
        str: The HTML code of a row.
    """
    pct = null_count / total_count
    return Template(
        """<tr>
    <th style="background-color: rgba(0, 191, 255, {{null_pct}})">{{column}}</th>
    <td {{num_style}}>{{null_pct}}</td>
    <td {{num_style}}>{{null_count}}</td>
    <td {{num_style}}>{{total_count}}</td>
</tr>"""
    ).render(
        {
            "num_style": f'style="text-align: right; background-color: rgba(0, 191, 255, {pct})"',
            "column": column,
            "null_count": f"{null_count:,}",
            "null_pct": f"{pct:.4f}",
            "total_count": f"{total_count:,}",
        }
    )


class TemporalNullValueSection(BaseSection):
    r"""Implements a section to analyze the temporal distribution of null
    values.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
    """

    def __init__(self, df: DataFrame, dt_column: str, period: str) -> None:
        self._df = df
        self._dt_column = dt_column
        self._period = period

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._dt_column,
                "figure": create_temporal_null_figure(
                    df=self._df, dt_column=self._dt_column, period=self._period
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the monthly distribution of null values.
The column {{column}} is used to define the month of each row.

{{figure}}
"""


def create_temporal_null_figure(
    df: DataFrame,
    dt_column: str,
    period: str,
    ncols: int = 2,
    figsize: tuple[int, int] = (700, 300),
) -> str:
    r"""Creates a HTML representation of a figure with the temporal null
    value distribution.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        ncols (int, optional): Specifies the number of columns.
            Default: ``2``
        figsize (``tuple``, optional): Specifies the individual figure
            size in pixels. The first dimension is the width and the
            second is the height.  Default: ``(700, 300)``

    Returns:
    -------
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""
    df = df.copy()
    columns = sorted([col for col in df.columns if col != dt_column])
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)

    nrows = math.ceil(len(columns) / ncols)
    fig = make_subplots(
        rows=nrows,
        cols=ncols,
        subplot_titles=columns,
        specs=[[{"secondary_y": True} for _ in range(ncols)] for _ in range(nrows)],
    )

    for i, column in enumerate(columns):
        x, y = i // ncols, i % ncols
        null_col = f"__{column}_isnull__"
        df2 = df[[column, dt_col]].copy()
        df2.loc[:, null_col] = df2.loc[:, column].isnull()

        df_sum = df2.groupby(dt_col)[null_col].sum().sort_index()
        df_count = df2.groupby(dt_col)[null_col].count().sort_index()
        labels = [str(dt) for dt in df_sum.index]

        fig.add_trace(
            go.Bar(
                x=labels,
                y=df_count.to_numpy(),
                marker=dict(color="rgba(0, 191, 255, 0.9)"),
            ),
            row=x + 1,
            col=y + 1,
            secondary_y=False,
        )
        fig.add_trace(
            go.Bar(
                x=labels,
                y=df_sum.to_numpy(),
                marker=dict(color="rgba(255, 191, 0, 0.9)"),
            ),
            row=x + 1,
            col=y + 1,
            secondary_y=False,
        )
        fig.add_trace(
            go.Scatter(
                x=labels,
                y=df_sum.to_numpy() / df_count.to_numpy(),
                marker=dict(color="rgba(0, 71, 171, 0.9)"),
            ),
            row=x + 1,
            col=y + 1,
            secondary_y=True,
        )

    fig.update_yaxes(
        title_text=(
            '<span style="color:RGB(255, 191, 0)">null</span>/'
            '<span style="color:RGB(0, 191, 255)">total</span>'
        ),
        secondary_y=False,
    )
    fig.update_yaxes(
        title_text='<span style="color:RGB(0, 71, 171)">percentage</span>',
        secondary_y=True,
    )
    fig.update_layout(
        height=figsize[1] * nrows + 120,
        width=figsize[0] * ncols,
        showlegend=False,
        barmode="overlay",
    )
    return plotly.io.to_html(fig, full_html=False)



from __future__ import annotations

__all__ = ["ColumnTypeAnalyzer"]

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnTypeSection
from flamme.utils.dtype import column_types


class ColumnTypeAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to find all the value types in each
    column.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnTypeAnalyzer
        >>> analyzer = ColumnTypeAnalyzer()
        >>> analyzer
        ColumnTypeAnalyzer()
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, df: DataFrame) -> ColumnTypeSection:
        return ColumnTypeSection(dtypes=df.dtypes.to_dict(), types=column_types(df))



from __future__ import annotations

__all__ = ["TemporalContinuousDistributionSection"]

from collections.abc import Sequence

import pandas as pd
import plotly
import plotly.express as px
from jinja2 import Template
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)


class TemporalContinuousDistributionSection(BaseSection):
    r"""Implements a section that analyzes the temporal distribution of a
    column with continuous values.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column of the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``
    """

    def __init__(
        self,
        df: DataFrame,
        column: str,
        dt_column: str,
        period: str,
        log_y: bool = False,
    ) -> None:
        self._df = df
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._log_y = log_y

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def log_y(self) -> bool:
        return self._log_y

    @property
    def period(self) -> str:
        return self._period

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    log_y=self._log_y,
                ),
                "table": create_temporal_table(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column {{column}} by using the column {{dt_column}}.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    df: DataFrame, column: str, dt_column: str, period: str, log_y: bool = False
) -> str:
    r"""Creates a HTML representation of a figure with the temporal value
    distribution.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``

    Returns:
    -------
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""
    df = df[[column, dt_column]].copy()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period).astype(str)

    fig = px.box(
        df,
        x=dt_col,
        y=column,
        title=f"Distribution of values for column {column}",
        labels={column: "value", dt_col: "time"},
        points="outliers",
        log_y=log_y,
    )
    return plotly.io.to_html(fig, full_html=False)


def create_temporal_table(df: DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a table with some statistics
    about the temporal value distribution.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
    -------
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return ""
    df = df[[column, dt_column]].copy()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)
    df_stats = (
        df.groupby(dt_col)[column]
        .agg(
            [
                "count",
                "mean",
                "median",
                "min",
                "max",
                "std",
                ("q01", lambda x: x.quantile(0.01)),
                ("q10", lambda x: x.quantile(0.1)),
                ("q25", lambda x: x.quantile(0.25)),
                ("q75", lambda x: x.quantile(0.75)),
                ("q90", lambda x: x.quantile(0.9)),
                ("q99", lambda x: x.quantile(0.99)),
            ]
        )
        .sort_index()
    )

    rows = []
    for row in df_stats.itertuples():
        rows.append(create_temporal_table_row(row))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period of column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>column</th>
                <th>count</th>
                <th>mean</th>
                <th>std</th>
                <th>min</th>
                <th>quantile 1%</th>
                <th>quantile 10%</th>
                <th>quantile 25%</th>
                <th>median</th>
                <th>quantile 75%</th>
                <th>quantile 90%</th>
                <th>quantile 99%</th>
                <th>max</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_table_row(row: pd.core.frame.Pandas) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
    ----
        row ("pd.core.frame.Pandas"): Specifies a DataFrame row.

    Returns:
    -------
        str: The HTML code of a row.
    """
    return Template(
        """<tr>
    <th>{{datetime}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{mean}}</td>
    <td {{num_style}}>{{std}}</td>
    <td {{num_style}}>{{min}}</td>
    <td {{num_style}}>{{q01}}</td>
    <td {{num_style}}>{{q10}}</td>
    <td {{num_style}}>{{q25}}</td>
    <td {{num_style}}>{{median}}</td>
    <td {{num_style}}>{{q75}}</td>
    <td {{num_style}}>{{q90}}</td>
    <td {{num_style}}>{{q99}}</td>
    <td {{num_style}}>{{max}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "datetime": row.Index,
            "count": f"{row.count:,}",
            "mean": f"{row.mean:,.4f}",
            "median": f"{row.median:,.4f}",
            "min": f"{row.min:,.4f}",
            "max": f"{row.max:,.4f}",
            "std": f"{row.std:,.4f}",
            "q01": f"{row.q01:,.4f}",
            "q10": f"{row.q10:,.4f}",
            "q25": f"{row.q25:,.4f}",
            "q75": f"{row.q75:,.4f}",
            "q90": f"{row.q90:,.4f}",
            "q99": f"{row.q99:,.4f}",
        }
    )



from __future__ import annotations

__all__ = ["TemporalNullValueAnalyzer", "NullValueAnalyzer"]

import logging

import numpy as np
from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import EmptySection
from flamme.section.null import NullValueSection, TemporalNullValueSection

logger = logging.getLogger(__name__)


class NullValueAnalyzer(BaseAnalyzer):
    r"""Implements a null value analyzer.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import NullValueAnalyzer
        >>> analyzer = NullValueAnalyzer()
        >>> analyzer
        NullValueAnalyzer()
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, df: DataFrame) -> NullValueSection:
        return NullValueSection(
            columns=list(df.columns),
            null_count=df.isnull().sum().to_frame("count")["count"].to_numpy(),
            total_count=np.full((df.shape[1],), df.shape[0]),
        )


class TemporalNullValueAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of null
    values.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import TemporalNullValueAnalyzer
        >>> analyzer = TemporalNullValueAnalyzer("datetime", period="M")
        >>> analyzer
        TemporalNullValueAnalyzer(dt_column=datetime, period=M)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...         "datetime": pd.to_datetime(
        ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, dt_column: str, period: str) -> None:
        self._dt_column = dt_column
        self._period = period

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(dt_column={self._dt_column}, period={self._period})"

    def analyze(self, df: DataFrame) -> TemporalNullValueSection | EmptySection:
        if self._dt_column not in df:
            logger.info(
                "Skipping monthly null value analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return TemporalNullValueSection(df=df, dt_column=self._dt_column, period=self._period)



from __future__ import annotations

__all__ = ["MappingAnalyzer"]

from collections.abc import Mapping

from coola.utils import str_indent, str_mapping
from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.section import SectionDict


class MappingAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to analyze multiple analyzers.

    Args:
    ----
        analyzers (``Mapping``): Specifies the mappings to analyze.
            The key of each analyzer is used to organize the metrics
            and report.
    """

    def __init__(self, analyzers: Mapping[str, BaseAnalyzer | dict]) -> None:
        self._analyzers = {name: setup_analyzer(analyzer) for name, analyzer in analyzers.items()}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self._analyzers))}\n)"

    def analyze(self, df: DataFrame) -> SectionDict:
        return SectionDict(
            {name: analyzer.analyze(df) for name, analyzer in self._analyzers.items()}
        )
